////////////////////////////////////////////////////////////////////////////////
// chaos_image_manager.cpp : image manager

////////////////////////////////////////////////////////////////////////////////
// include

#include "stdafx.h"

#include "chaos_image_manager.h"
#include "chaos_file_serializer.h"

////////////////////////////////////////////////////////////////////////////////
// chaos_image_manager : image manager class

////////////////////////////////////////////////////////////////////////////////
// create & destroy

// default constructor

chaos_image_manager::chaos_image_manager() :
	screen_dim(0, 0),
	main_image(NULL),
	prompt_image(NULL),
	screen_image(NULL),
	stretch_blt_info(NULL),
	image_realizer(NULL),
	message_imager(NULL)
{
	main_image = new chaos_image;
	prompt_image = new chaos_image;
	screen_image = new chaos_image;
	stretch_blt_info = new chaos_image::fast_smooth_stretch_blt_info();
	image_realizer = new chaos_image_realizer;
	message_imager = new chaos_message_imager;
}

// destructor

chaos_image_manager::~chaos_image_manager() {
	dispose();
	delete main_image;
	delete prompt_image;
	delete screen_image;
	delete stretch_blt_info;
	delete image_realizer;
	delete message_imager;
}

// initialize

bool chaos_image_manager::initialize(
	chaos_dim sys_dim,
	const chaos_string& font_name,
	int font_size,
	bool font_bold,
	bool font_anti_alias,
	int image_buffer_count)
{
	screen_dim = sys_dim;
	if (!main_image->is_valid()) {
		if (!main_image->create(sys_dim)) {
			return false;
		}
	}
	if (!prompt_image->is_valid()) {
		if (!prompt_image->create(32, 32)) {
			return false;
		}
	}
	if (!message_imager->initialize(sys_dim)) {
		return false;
	}
	message_imager->set_font_name(font_name);
	message_imager->set_font_size(font_size);
	message_imager->set_font_bold(font_bold);
	message_imager->set_font_anti_alias(font_anti_alias);
	if (!image_realizer->initialize(image_buffer_count, sys_dim)) {
		return false;
	}
	chaos_image sysetc_image;
	if (image_realizer->realize_image(
			&sysetc_image,
			chaos_image_realizer::encode_load_info("./sysetc.bmp")))
	{
		int color_x = 64, color_y = 0,
			alpha_x = 96, alpha_y = 0;
		for (int y = 0; y < prompt_image->get_height(); y++) {
			for (int x = 0; x < prompt_image->get_width(); x++) {
				chaos_uint32 col1 =
					sysetc_image.get_pixel(color_x+x, color_y+y),
						col2 =
					sysetc_image.get_pixel(alpha_x+x, alpha_y+y);
				int red = CHAOS_RGBA_RED(col1),
					green = CHAOS_RGBA_GREEN(col1),
					blue = CHAOS_RGBA_BLUE(col1),
					alpha = (CHAOS_RGBA_RED(col2)+
						CHAOS_RGBA_GREEN(col2)+
						CHAOS_RGBA_BLUE(col2))/3;
				prompt_image->set_pixel(
					x, y,
					CHAOS_RGBA(red, green, blue, alpha));
			}
		}
	}
	return true;
}

// dispose

void chaos_image_manager::dispose() {
	image_realizer->dispose();
	message_imager->dispose();
}

// change image buffer count

bool chaos_image_manager::change_image_buffer_count(
	int image_buffer_count)
{
	bool result = true;
	if (image_buffer_count != image_realizer->get_reserve_count()) {
		result = image_realizer->initialize(
			image_buffer_count,
			image_realizer->get_sys_dim());
	}
	return result;
}

// change font

bool chaos_image_manager::change_font(
	const chaos_string& font_name,
	int font_size,
	bool font_bold,
	bool font_anti_alias)
{
	if (font_name != message_imager->get_font_name()) {
		message_imager->set_font_name(font_name);
	}
	if (font_size != message_imager->get_font_size()) {
		message_imager->set_font_size(font_size);
	}
	if (font_bold != message_imager->is_font_bold()) {
		message_imager->set_font_bold(font_bold);
	}
	if (font_anti_alias != message_imager->is_font_anti_alias()) {
		message_imager->set_font_anti_alias(font_anti_alias);
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// draw

// draw main image

void chaos_image_manager::draw_main_image(
	const chaos_string& info,
	const chaos_rect* sel_rect,
	chaos_color sel_color,
	bool prompt_visible, bool prompt_place,
	chaos_pos prompt_place_pos,
	chaos_rect* update_rect,
	const chaos_rect* clip_rect)
{
	ASSERT(main_image->is_valid());
	chaos_rect main_image_rect(
		chaos_pos(0, 0), main_image->get_dim());
	chaos_pos base_pos =
		chaos_image_realizer::extract_base_position(info);
	bool clipping = false;
	chaos_rect real_clip_rect;
	if (clip_rect != NULL) {
		clipping = true;
		real_clip_rect = *clip_rect+base_pos;
		real_clip_rect &= main_image_rect;
	}
	// background image
	image_realizer->realize_image(
		main_image,
		info,
		clipping? &real_clip_rect: NULL);
	chaos_rect background_rect = main_image_rect;
	if (clipping) {
		background_rect &= real_clip_rect;
	}
	chaos_rect real_update_rect = background_rect;
	// selection
	chaos_rect sel_fill_rect = *sel_rect+base_pos;
	if (clipping) {
		sel_fill_rect &= real_clip_rect;
	}
	if (!sel_fill_rect.empty()) {
		main_image->trans_fill(
			sel_color.get_red(),
			sel_color.get_green(),
			sel_color.get_blue(),
			&sel_fill_rect);
		real_update_rect |= sel_fill_rect;
	}
	// message image
	chaos_rect message_valid_rect =
		message_imager->get_message_valid_rect()+base_pos;
	if (clipping) {
		message_valid_rect &= real_clip_rect;
	}
	if (!message_valid_rect.empty()) {
		if (message_imager->is_font_anti_alias()) {
			main_image->alpha_blt(
				message_imager->get_message_image(),
				base_pos,
				&message_valid_rect);
		} else {
			main_image->mask_blt(
				message_imager->get_message_image(),
				base_pos,
				&message_valid_rect);
		}
		real_update_rect |= message_valid_rect;
	}
	// prompt
	if (prompt_visible) {
		ASSERT(prompt_image->is_valid());
		chaos_pos prompt_pos = base_pos;
		if (prompt_place) {
			prompt_pos += prompt_place_pos;
		} else {
			prompt_pos += message_imager->get_message_pos();
		}
		chaos_rect prompt_rect(
			prompt_pos,
			prompt_image->get_dim());
		if (clipping) {
			prompt_rect &= real_clip_rect;
		}
		main_image->alpha_blt(
			prompt_image,
			prompt_pos,
			&prompt_rect);
		real_update_rect |= prompt_rect;
	}
	real_update_rect &= main_image_rect;
	if (update_rect != NULL) {
		*update_rect = real_update_rect;
	}
	if (screen_image->is_valid()) {
		chaos_rect screen_clip_rect =
			map_coordinate(
				chaos_pos(),
				screen_dim,
				chaos_pos(),
				main_image->get_dim(),
				real_update_rect,
				true);
		screen_image->fast_smooth_stretch_blt(
			stretch_blt_info,
			main_image,
			chaos_pos(0, 0),
			screen_dim,
			&screen_clip_rect);
	}
}

////////////////////////////////////////////////////////////////////////////////
// message

// update message image

void chaos_image_manager::update_message_image(
	const chaos_string& message_text,
	chaos_rect* update_rect)
{
	message_imager->update_message_text(
		message_text,
		update_rect);
}

// clear message image

void chaos_image_manager::clear_message_image() {
	message_imager->clear_message_text();
}

////////////////////////////////////////////////////////////////////////////////
// coordinate

// map coordinate

chaos_pos chaos_image_manager::map_coordinate(
	chaos_pos dst_ofs,
	chaos_dim dst_dim,
	chaos_pos src_ofs,
	chaos_dim src_dim,
	chaos_pos src_pos,
	bool ceil)
{
	chaos_pos dst_pos = src_pos;
	dst_pos -= src_ofs;
	if (src_dim.empty() ||
		(dst_dim == src_dim))
	{
		// nothing
	} else if (!ceil) {
		dst_pos.x =
			(dst_pos.x*dst_dim.width)/src_dim.width;
		dst_pos.y =
			(dst_pos.y*dst_dim.height)/src_dim.height;
	} else {
		dst_pos.x =
			(dst_pos.x*dst_dim.width+src_dim.width-1)/src_dim.width;
		dst_pos.y =
			(dst_pos.y*dst_dim.height+src_dim.height-1)/src_dim.height;
	}
	dst_pos += dst_ofs;
	return dst_pos;
}

chaos_rect chaos_image_manager::map_coordinate(
	chaos_pos dst_ofs,
	chaos_dim dst_dim,
	chaos_pos src_ofs,
	chaos_dim src_dim,
	chaos_rect src_rect,
	bool outside)
{
	chaos_rect dst_rect = src_rect;
	if (src_dim.empty() ||
		(dst_dim == src_dim))
	{
		dst_rect -= src_ofs;
		dst_rect += dst_ofs;
	} else {
		if (outside) {
			dst_rect.x--;
			dst_rect.y--;
			dst_rect.width += 2;
			dst_rect.height += 2;
		}
		dst_rect = chaos_rect(
			map_coordinate(
				dst_ofs,
				dst_dim,
				src_ofs,
				src_dim,
				dst_rect.get_lt_pos(),
				false),
			map_coordinate(
				dst_ofs,
				dst_dim,
				src_ofs,
				src_dim,
				dst_rect.get_rb_pos(),
				true));
		if (outside) {
			dst_rect.x--;
			dst_rect.y--;
			dst_rect.width += 2;
			dst_rect.height += 2;
		}
	}
	return dst_rect;
}

// map coordinate system to screen

chaos_pos chaos_image_manager::map_sys2screen(
	chaos_pos sys_pos,
	bool ceil)
{
	return map_coordinate(
		screen_ofs,
		screen_dim,
		chaos_pos(),
		main_image->get_dim(),
		sys_pos,
		ceil);
}

chaos_rect chaos_image_manager::map_sys2screen(
	chaos_rect sys_rect,
	bool outside)
{
	return map_coordinate(
		screen_ofs,
		screen_dim,
		chaos_pos(),
		main_image->get_dim(),
		sys_rect,
		outside);
}

// map coordinate screen to system

chaos_pos chaos_image_manager::map_screen2sys(
	chaos_pos screen_pos,
	bool ceil)
{
	return map_coordinate(
		chaos_pos(),
		main_image->get_dim(),
		screen_ofs,
		screen_dim,
		screen_pos,
		ceil);
}

chaos_rect chaos_image_manager::map_screen2sys(
	chaos_rect screen_rect,
	bool outside)
{
	return map_coordinate(
		chaos_pos(),
		main_image->get_dim(),
		screen_ofs,
		screen_dim,
		screen_rect,
		outside);
}

////////////////////////////////////////////////////////////////////////////////
// stretch

// change screen dim

void chaos_image_manager::set_frame_rect(
	chaos_rect frame_rect)
{
	chaos_dim frame_dim(
			frame_rect.get_dim()),
		new_screen_dim(
			frame_dim.inside_dim(main_image->get_dim()));
	chaos_pos new_screen_ofs(
			(frame_dim.width-new_screen_dim.width)/2,
			(frame_dim.height-new_screen_dim.height)/2);
	new_screen_ofs += frame_rect.get_lt_pos();
	screen_ofs = new_screen_ofs;
	if (new_screen_dim != screen_dim) {
		if (screen_image->is_valid()) {
			screen_image->destroy();
		}
		screen_dim = new_screen_dim;
	}
}

// get screen image

const chaos_image* chaos_image_manager::get_screen_image() {
	const chaos_image* image = get_main_image();
	if (image->is_valid() &&
		(image->get_dim() != screen_dim) &&
		!screen_dim.empty())
	{
		if (!screen_image->is_valid()) {
			screen_image->create(screen_dim, false);
			screen_image->fast_smooth_stretch_blt(
				stretch_blt_info,
				image,
				chaos_pos(0, 0),
				screen_dim);
		}
		image = screen_image;
	}
	return image;
}
